import pytest

hardware = "daq2"
classname = "adi.ad9144"


@pytest.mark.parametrize("classname, hardware", [(classname, hardware)])
@pytest.mark.parametrize("channel", [0, 1, [0, 1]])
def test_ad9144_tx_data(test_dma_tx, classname, hardware, channel):
    test_dma_tx(classname, hardware, channel)



# Copyright (C) 2019 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import print_function

import test.iio_scanner as iio_scanner
import unittest

import iio

import numpy as np
from adi import DAQ2

URI = "ip:analog"
dev_checked = False
found_dev = False


def check_dev(name):
    global dev_checked
    global found_dev
    if not dev_checked:
        found_dev, board = iio_scanner.find_device(name)
        if found_dev:
            global URI
            URI = board.uri
        dev_checked = True
    return found_dev


class TestDAQ2(unittest.TestCase):

    # do_plots = False

    def freq_est(self, y, fs):
        N = len(y)
        T = 1.0 / fs
        yf = np.fft.fft(y)
        yf = np.fft.fftshift(yf)
        xf = np.linspace(-1.0 / (2.0 * T), 1.0 / (2.0 * T), N)
        # if self.do_plots:
        #     import matplotlib.pyplot as plt
        #
        #     fig, ax = plt.subplots()
        #     ax.plot(xf, 2.0 / N * np.abs(yf))
        #     plt.show()
        indx = np.argmax(np.abs(yf))
        return np.abs(xf[indx])

    def setUp(self):
        pass

    def tearDown(self):
        pass

    @unittest.skipUnless(check_dev("daq2"), "daq2 not attached")
    def testDAQ2ADC(self):
        global URI
        # See if we can get non-zero data from ADC
        adc = DAQ2(uri=URI)
        adc.rx_enabled_channels = [0]
        data = adc.rx()
        s = np.sum(np.abs(data))
        del adc
        self.assertGreater(s, 0, "check non-zero data")

    @unittest.skipUnless(check_dev("daq2"), "daq2 not attached")
    def testDAQ2ADC_p2(self):
        global URI
        # See if we can get non-zero data from ADC
        adc = DAQ2(uri=URI)
        adc.rx_enabled_channels = [1]
        data = adc.rx()
        s = np.sum(np.abs(data))
        del adc
        self.assertGreater(s, 0, "check non-zero data")

    @unittest.skipUnless(check_dev("daq2"), "daq2 not attached")
    def testDAQ2ADC_dual(self):
        global URI
        # See if we can get non-zero data from ADC
        adc = DAQ2(uri=URI)
        adc.rx_enabled_channels = [0, 1]
        data = adc.rx()
        del adc
        s = np.sum(np.abs(data[0]))
        self.assertGreater(s, 0, "check non-zero data")
        s = np.sum(np.abs(data[1]))
        self.assertGreater(s, 0, "check non-zero data")

    @unittest.skipUnless(check_dev("daq2"), "daq2 not attached")
    def testDAQ2ADC_dual_real(self):
        global URI
        # See if we can get non-zero data from ADC
        adc = DAQ2(uri=URI)
        adc.rx_enabled_channels = [0, 1]
        data = adc.rx()
        del adc
        is_complex = False
        for d in data[0]:
            is_complex = is_complex or np.iscomplex(d)
        for d in data[1]:
            is_complex = is_complex or np.iscomplex(d)

        self.assertEqual(is_complex, False, "check real data")

    @unittest.skipUnless(check_dev("daq2"), "daq2 not attached")
    def testDAQ2DAC(self):
        global URI
        # See if we tx data from DAC
        dac = DAQ2(uri=URI)
        TXFS = dac.sample_rate
        dac.tx_enabled_channels = [0]
        N = 2 ** 15
        ts = 1 / float(TXFS)
        t = np.arange(0, N * ts, ts)
        fc = 10000
        d = np.cos(2 * np.pi * t * fc) * 2 ** 15 * 0.5
        dac.tx(d)
        del dac
        self.assertEqual(True, True, "transmit data failed")

    @unittest.skipUnless(check_dev("daq2"), "daq2 not attached")
    def testDAQ2DAC_p2(self):
        global URI
        # See if we tx data from DAC
        dac = DAQ2(uri=URI)
        TXFS = dac.sample_rate
        dac.tx_enabled_channels = [1]
        N = 2 ** 15
        ts = 1 / float(TXFS)
        t = np.arange(0, N * ts, ts)
        fc = 10000
        d = np.cos(2 * np.pi * t * fc) * 2 ** 15 * 0.5
        dac.tx(d)
        del dac
        self.assertEqual(True, True, "transmit data failed")

    @unittest.skipUnless(check_dev("daq2"), "daq2 not attached")
    def testDAQ2DAC_dual(self):
        global URI
        # See if we tx data from DAC
        dac = DAQ2(uri=URI)
        TXFS = dac.sample_rate
        dac.tx_enabled_channels = [0, 1]
        N = 2 ** 15
        ts = 1 / float(TXFS)
        t = np.arange(0, N * ts, ts)
        fc = 10000
        d = np.cos(2 * np.pi * t * fc) * 2 ** 15 * 0.5
        dac.tx([d, d])
        del dac
        self.assertEqual(True, True, "transmit data failed")

    @unittest.skipUnless(check_dev("daq2"), "daq2 not attached")
    def testDAQ2DMA(self):
        global URI
        # Test DMA
        daq = DAQ2(uri=URI)
        daq._ctx.set_timeout = 1000
        daq.tx_cyclic_buffer = True
        daq.tx_enabled_channels = [0, 1]
        # Create signal
        fs = int(daq.sample_rate)
        fc1 = 30 * 1e6
        fc2 = 100 * 1e6
        N = 2 ** 12
        ts = 1 / float(fs)
        t = np.arange(0, N * ts, ts)
        f1 = np.cos(2 * np.pi * t * fc1) * 2 ** 14
        f2 = np.cos(2 * np.pi * t * fc2) * 2 ** 14

        daq.rx_buffer_size = len(f1) * 2

        daq.tx([f1, f2])
        # Flush buffers
        for _ in range(10):
            data = daq.rx()
        del daq
        # Estimate frequency
        fc1_est = self.freq_est(data[0], fs)
        fc2_est = self.freq_est(data[1], fs)
        # Check Data
        diff = np.abs(fc1_est - fc1)
        self.assertGreater(fc1 * 0.01, diff, "Frequency offset TX1")
        diff = np.abs(fc2_est - fc2)
        self.assertGreater(fc2 * 0.01, diff, "Frequency offset TX2")

    @unittest.skipUnless(check_dev("daq2"), "daq2 not attached")
    def testDAQ2DDS(self):
        global URI
        # Test DMA
        daq = DAQ2(uri=URI)
        daq.tx_cyclic_buffer = True
        daq.tx_enabled_channels = [0, 1]
        daq.rx_buffer_size = 2 ** 14
        # Enable DDSs
        fs = int(daq.sample_rate)
        fc1 = 20 * 1e6
        fc2 = 70 * 1e6
        daq.dds_enabled = [1, 1, 1, 1, 1, 1, 1, 1]
        daq.dds_frequencies = [fc1, 0, fc2, 0, 0, 0, 0, 0]
        daq.dds_scales = [1, 0, 1, 0, 0, 0, 0, 0]
        daq.dds_phases = [0, 0, 0, 0, 0, 0, 0, 0]

        # Flush buffers
        for _ in range(10):
            data = daq.rx()
        del daq

        # Estimate frequency
        fc1_est = self.freq_est(data[0], fs)
        fc2_est = self.freq_est(data[1], fs)

        # Check Data
        diff = np.abs(fc1_est - fc1)
        self.assertGreater(fc1 * 0.01, diff, "Frequency offset TX1")
        diff = np.abs(fc2_est - fc2)
        self.assertGreater(fc2 * 0.01, diff, "Frequency offset TX2")

    @unittest.skipUnless(check_dev("daq2"), "daq2 not attached")
    def testDAQ2DDS_SimpleAPI(self):
        global URI
        # Test DMA
        daq = DAQ2(uri=URI)
        daq.tx_cyclic_buffer = True
        daq.tx_enabled_channels = [0, 1]
        daq.rx_buffer_size = 2 ** 14
        fs = int(daq.sample_rate)

        # Enable DDSs
        fc1 = 200000000
        daq.dds_single_tone(fc1, 0.5)

        # Flush buffers
        for _ in range(10):
            data = daq.rx()

        # Estimate frequency
        fc1_est = self.freq_est(data[0], fs)

        # Change DDS channel
        fc2 = 200000000
        daq.dds_single_tone(fc2, 0.5, 1)
        # Flush buffers
        for _ in range(10):
            data = daq.rx()
        del daq

        fc2_est = self.freq_est(data[1], fs)

        # Check Data
        diff = np.abs(fc1_est - fc1)
        self.assertGreater(fc1 * 0.01, diff, "Frequency offset TX1")
        diff = np.abs(fc2_est - fc2)
        self.assertGreater(fc2 * 0.01, diff, "Frequency offset TX2")


if __name__ == "__main__":
    unittest.main()



import numpy as np


def measure_phase_and_delay(chan0, chan1):
    """
    Measure phase and sample delay between two complex signals.
    Signals are assumed to have a single cross-correlation
    peak. Using noise sources are generally useful for this
    measurement and not sinusoids

    Parameters
    ----------
    chan0 : np.array complex
        Single complex signal
    chan1 : np.array complex
        Signel complex signal

    Returns
    -------
    (phase, sample_delay) : tuple(float,int)
        phase: phase difference between signals in degrees
        sample_delay: sample offset difference between signals
    """
    cor = np.correlate(chan0, chan1, "full")
    i = np.argmax(cor)
    m = cor[i]
    sample_delay = len(chan0) - i - 1
    return (np.angle(m) * 180 / np.pi, sample_delay)


def measure_phase(chan0, chan1):
    """
    Measure phase between two complex signals.
    Instanteous phase at each point between signals
    is measured then averaged. This measurement assumes
    no sample offset between signals

    Parameters
    ----------
    chan0 : np.array complex
        Single complex signal
    chan1 : np.array complex
        Signel complex signal

    Returns
    -------
    phase : float
        Phase difference between signals in degrees
    """
    errorV = np.angle(chan0 * np.conj(chan1)) * 180 / np.pi
    error = np.mean(errorV)
    return error



from test.rf.phase import *

# import matplotlib.pyplot as plt
import pytest
from numpy import abs, exp, linspace, pi, real


def test_measure_phase():
    fs = 64
    length = 600  # seconds
    N = fs * length
    t = linspace(0, length, num=N, endpoint=False)

    # Generate a sinusoid at frequency f
    f = 10  # Hz
    a = exp(1j * 2 * pi * f * t)
    b = exp(1j * (2 * pi * f * t - pi / 4))
    # plt.plot(t, np.real(a))
    # plt.plot(t, np.real(b))
    # plt.show()

    phase = measure_phase(a, b)

    assert abs(phase - 45) < 0.01


def test_measure_phase_and_delay_delay():
    buff_size = 1024
    B = 2 ** 15 * 0.8
    a = np.random.uniform(-B, B, buff_size * 4)
    b = a * exp(1j * (-pi / 4))
    a = a[1:]
    b = b[:-1]
    # plt.plot(t, np.real(a))
    # plt.plot(t, np.real(b))
    # plt.show()
    phase, delay = measure_phase_and_delay(a, b)
    assert delay == 1
    assert abs(phase - 45) < 0.01


def test_measure_phase_and_delay_phase():
    buff_size = 1024
    B = 2 ** 15 * 0.8
    a = np.random.uniform(-B, B, buff_size * 4)
    b = a * exp(1j * (-pi / 4))
    # plt.plot(t, np.real(a))
    # plt.plot(t, np.real(b))
    # plt.show()
    phase, delay = measure_phase_and_delay(a, b)
    assert delay == 0
    assert abs(phase - 45) < 0.01



import pytest

hardware = "rpi"
classname = "adi.adis16507"


@pytest.mark.parametrize(
    "classname, hardware, attr, max_pow, tol",
    [(classname, hardware, "sample_rate", 11, 0)],
)
def test_adis16507_sample_rate(
    test_attribute_single_value_pow2, classname, hardware, attr, max_pow, tol
):
    test_attribute_single_value_pow2(classname, hardware, attr, max_pow, tol)


@pytest.mark.parametrize("classname, hardware", [(classname, hardware)])
@pytest.mark.parametrize("channel", [0, 1, 2, 3, 4, 5])
def test_adis16507_rx_data(test_dma_rx, classname, hardware, channel):
    test_dma_rx(classname, hardware, channel)



# Copyright (C) 2019 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from adi.context_manager import context_manager
from adi.rx_tx import rx_tx


class adrv9009(rx_tx, context_manager):
    """ ADRV9009 Transceiver """

    _complex_data = True
    _rx_channel_names = ["voltage0_i", "voltage0_q", "voltage1_i", "voltage1_q"]
    _tx_channel_names = ["voltage0", "voltage1", "voltage2", "voltage3"]
    _device_name = ""

    def __init__(self, uri=""):

        context_manager.__init__(self, uri, self._device_name)

        self._ctrl = self._ctx.find_device("adrv9009-phy")
        self._rxadc = self._ctx.find_device("axi-adrv9009-rx-hpc")
        self._rxobs = self._ctx.find_device("axi-adrv9009-rx-obs-hpc")
        self._txdac = self._ctx.find_device("axi-adrv9009-tx-hpc")
        self._ctx.set_timeout(30000)  # Needed for loading profiles

        rx_tx.__init__(self)

    @property
    def profile(self):
        """Load profile file. Provide path to profile file to attribute"""
        return self._get_iio_dev_attr("profile_config")

    @profile.setter
    def profile(self, value):
        with open(value, "r") as file:
            data = file.read()
        self._set_iio_dev_attr_str("profile_config", data)

    @property
    def calibrate_rx_phase_correction_en(self):
        """calibrate_rx_phase_correction_en: Enable RX Phase Correction Calibration"""
        return self._get_iio_dev_attr("calibrate_rx_phase_correction_en")

    @calibrate_rx_phase_correction_en.setter
    def calibrate_rx_phase_correction_en(self, value):
        self._set_iio_dev_attr_str("calibrate_rx_phase_correction_en", value)

    @property
    def calibrate_rx_qec_en(self):
        """calibrate_rx_qec_en: Enable RX QEC Calibration"""
        return self._get_iio_dev_attr("calibrate_rx_qec_en")

    @calibrate_rx_qec_en.setter
    def calibrate_rx_qec_en(self, value):
        self._set_iio_dev_attr_str("calibrate_rx_qec_en", value)

    @property
    def calibrate_tx_qec_en(self):
        """calibrate_tx_qec_en: Enable TX QEC Calibration"""
        return self._get_iio_dev_attr("calibrate_tx_qec_en")

    @calibrate_tx_qec_en.setter
    def calibrate_tx_qec_en(self, value):
        self._set_iio_dev_attr_str("calibrate_tx_qec_en", value)

    @property
    def calibrate(self):
        """calibrate: Trigger Calibration"""
        return self._get_iio_dev_attr("calibrate")

    @calibrate.setter
    def calibrate(self, value):
        self._set_iio_dev_attr_str("calibrate", value)

    @property
    def gain_control_mode_chan0(self):
        """gain_control_mode_chan0: Mode of receive path AGC. Options are:
        slow_attack, manual"""
        return self._get_iio_attr_str("voltage0", "gain_control_mode", False)

    @gain_control_mode_chan0.setter
    def gain_control_mode_chan0(self, value):
        self._set_iio_attr("voltage0", "gain_control_mode", False, value)

    @property
    def gain_control_mode_chan1(self):
        """gain_control_mode_chan1: Mode of receive path AGC. Options are:
        slow_attack, manual"""
        return self._get_iio_attr_str("voltage1", "gain_control_mode", False)

    @gain_control_mode_chan1.setter
    def gain_control_mode_chan1(self, value):
        self._set_iio_attr("voltage1", "gain_control_mode", False, value)

    @property
    def rx_hardwaregain_chan0(self):
        """rx_hardwaregain: Gain applied to RX path channel 0. Only applicable when
        gain_control_mode is set to 'manual'"""
        return self._get_iio_attr("voltage0", "hardwaregain", False)

    @rx_hardwaregain_chan0.setter
    def rx_hardwaregain_chan0(self, value):
        if self.gain_control_mode_chan0 == "manual":
            self._set_iio_attr("voltage0", "hardwaregain", False, value)

    @property
    def rx_hardwaregain_chan1(self):
        """rx_hardwaregain: Gain applied to RX path channel 1. Only applicable when
        gain_control_mode is set to 'manual'"""
        return self._get_iio_attr("voltage1", "hardwaregain", False)

    @rx_hardwaregain_chan1.setter
    def rx_hardwaregain_chan1(self, value):
        if self.gain_control_mode_chan1 == "manual":
            self._set_iio_attr("voltage1", "hardwaregain", False, value)

    @property
    def tx_hardwaregain_chan0(self):
        """tx_hardwaregain: Attenuation applied to TX path channel 0"""
        return self._get_iio_attr("voltage0", "hardwaregain", True)

    @tx_hardwaregain_chan0.setter
    def tx_hardwaregain_chan0(self, value):
        self._set_iio_attr("voltage0", "hardwaregain", True, value)

    @property
    def tx_hardwaregain_chan1(self):
        """tx_hardwaregain: Attenuation applied to TX path channel 1"""
        return self._get_iio_attr("voltage1", "hardwaregain", True)

    @tx_hardwaregain_chan1.setter
    def tx_hardwaregain_chan1(self, value):
        self._set_iio_attr("voltage1", "hardwaregain", True, value)

    @property
    def rx_rf_bandwidth(self):
        """rx_rf_bandwidth: Bandwidth of front-end analog filter of RX path"""
        return self._get_iio_attr("voltage0", "rf_bandwidth", False)

    @property
    def tx_rf_bandwidth(self):
        """tx_rf_bandwidth: Bandwidth of front-end analog filter of TX path"""
        return self._get_iio_attr("voltage0", "rf_bandwidth", True)

    @property
    def rx_sample_rate(self):
        """rx_sample_rate: Sample rate RX path in samples per second"""
        return self._get_iio_attr("voltage0", "sampling_frequency", False)

    @property
    def tx_sample_rate(self):
        """tx_sample_rate: Sample rate TX path in samples per second"""
        return self._get_iio_attr("voltage0", "sampling_frequency", True)

    @property
    def trx_lo(self):
        """trx_lo: Carrier frequency of TX and RX path"""
        return self._get_iio_attr("altvoltage0", "frequency", True)

    @trx_lo.setter
    def trx_lo(self, value):
        self._set_iio_attr("altvoltage0", "frequency", True, value)



import os
import socket
import subprocess
import sys

import iio

import scapy.all as scapy

HOSTNAMES = [
    "analog.local",
    "analog.lan",
    "analog-zc706-daq2.local",
    "analog-zc706-fmcomms2.local",
]


def check_iio(address):
    try:
        return iio.Context("ip:" + address)
    except:
        return False


def check_iio_uri(uri):
    try:
        return iio.Context(uri)
    except:
        return False


def dump(obj):
    for attr in dir(obj):
        print("obj.%s = %r" % (attr, getattr(obj, attr)))


def check_exist(address):
    arp_request = scapy.ARP(pdst=address)
    broadcast = scapy.Ether(dst="ff:ff:ff:ff:ff:ff")
    arp_request_broadcast = broadcast / arp_request
    answered_list = scapy.srp(arp_request_broadcast, timeout=1, verbose=False)[0]
    clients_list = []
    for element in answered_list:
        client_dict = {"ip": element[1].psrc, "mac": element[1].hwsrc}
        clients_list.append(client_dict)
    return clients_list


class device:
    def __init__(self, name, channels=[]):
        self.name = name
        self.channels = channels


class board:
    def __init__(self, name, uri):
        self.name = name
        self.uri = uri


def check_config(ctx, devices):
    found = 0
    try:
        for dev in ctx.devices:
            for fdev in devices:
                if not dev.name:
                    continue
                if dev.name.lower() == fdev.name.lower():
                    if fdev.channels:
                        chans = 0
                        for chan in dev.channels:
                            chans = chans + chan.scan_element
                        found = found + (chans == fdev.channels)
                    else:
                        found = found + 1
        return found == len(devices)
    except:
        return False


def check_board_other(ctx):
    if check_config(
        ctx, [device("ad7291-ccbox"), device("ad9361-phy"), device("cf-ad9361-lpc", 4)]
    ):
        return "packrf"
    if check_config(
        ctx, [device("ad9517"), device("ad9361-phy"), device("cf-ad9361-lpc", 4)]
    ):
        return "adrv9361"
    if check_config(ctx, [device("ad7291-bob"), device("cf-ad9361-lpc", 2)]):
        return "adrv9364"
    if check_config(
        ctx, [device("adm1177"), device("ad9361-phy"), device("cf-ad9361-lpc", 2)]
    ):
        return "pluto"
    if check_config(
        ctx, [device("ad7291"), device("ad9361-phy"), device("cf-ad9361-lpc", 4)]
    ):
        return "fmcomms2"
    if check_config(
        ctx, [device("ad7291"), device("ad9361-phy"), device("cf-ad9361-lpc", 2),],
    ):
        return "fmcomms4"
    if check_config(ctx, [device("ad9361-phy"), device("ad9361-phy-b")]):
        return "fmcomms5"
    if check_config(ctx, [device("ad9361-phy"), device("cf-ad9361-lpc", 2)]):
        return "ad9364"
    if check_config(ctx, [device("ad9361-phy"), device("cf-ad9361-lpc", 4)]):
        return "ad9361"

    if check_config(ctx, [device("axi-ad9144-hpc", 4), device("axi-ad9680-hpc", 2)]):
        return "daq2"

    if check_config(ctx, [device("axi-ad9152-hpc", 2), device("axi-ad9680-hpc", 2)]):
        return "daq3"

    if check_config(ctx, [device("adrv9009-phy"), device("adrv9009-phy-b")]):
        return "adrv9009-dual"
    if check_config(ctx, [device("adrv9009-phy")]):
        return "adrv9009"

    if check_config(ctx, [device("ad9371-phy")]):
        return "ad9371"

    for dev in ctx.devices:
        if dev.name:
            return dev.name


def ip_scan_auto():
    boards = []
    addresses = []
    for host in HOSTNAMES:
        try:
            for ip in socket.gethostbyname_ex(host):
                if type(ip) == list and ip:
                    if ip[0] not in addresses:
                        addresses.append(ip[0])
        except:
            continue
    for address in addresses:
        ctx = check_iio(address)
        if ctx:
            name = check_board_other(ctx)
            b = board(name, "ip:" + address)
            if b not in boards:
                boards.append(b)
    return boards


def ip_scan(subnet):
    boards = []
    for ping in range(0, 255):
        address = subnet + "." + str(ping)
        res = check_exist(address)
        if res:
            ctx = check_iio(address)
            if ctx:
                print("ping to", address, "OK MAC", res[0]["mac"])
                name = check_board_other(ctx)
                boards.append(board(name, "ip:" + address))
    return boards


def scan_all(skip_usb=False):
    boards = []

    # FIND IP
    bs = ip_scan_auto()
    # bs = ip_scan("192.168.86")

    if bs not in boards:
        boards = boards + bs

    # Find USB/LOCAL
    if not skip_usb:
        ctxs = iio.scan_contexts()
        for ctx in ctxs:
            c = iio.Context(ctx)
            name = check_board_other(c)
            if name:
                if c.name == "local":
                    boards.append(board(name, "local:"))
                else:
                    boards.append(board(name, ctx))
    return boards


def get_device(uri):
    ctx = check_iio_uri(uri)
    if ctx:
        board_name = check_board_other(ctx)
        if board_name:
            if ctx.name == "local":
                b = board(board_name, "local:")
            else:
                b = board(board_name, uri)
        return b
    return None


def find_device(names, uri=None):
    skip_usb = False
    if not isinstance(names, list):
        names = [names]
    if uri:
        b = get_device(uri)
        if b:
            boards = [b]
        else:
            return (False, [])
    else:
        boards = scan_all(skip_usb)

    for b in boards:
        for name in names:
            if b.name == name:
                return (True, b)

    return (False, [])


if __name__ == "__main__":
    bs = scan_all()
    for b in bs:
        print(b.uri)
        print(b.name)
        print("---------")



from test.rf.spec import *

import numpy as np
import pytest


def test_spec_est_t1():
    fs = 64
    length = 600  # seconds
    N = fs * length
    t = linspace(0, length, num=N, endpoint=False)

    # Generate a sinusoid at frequency f
    f = 10  # Hz
    a = cos(2 * pi * f * t) * 2 ** 15

    amp, freqs = spec_est(a, fs, ref=2 ** 15, plot=False)

    assert np.abs(max(amp) + 6) < 0.03


def test_spec_est_t2():
    fs = 64
    length = 600  # seconds
    N = fs * length
    t = linspace(0, length, num=N, endpoint=False)

    # Generate a sinusoid at frequency f
    f = 10  # Hz
    a = exp(1j * 2 * pi * f * t) * 2 ** 15

    amp, freqs = spec_est(a, fs, ref=2 ** 15, plot=False)

    assert np.abs(max(amp)) < 0.03


def test_spec_est_t3():
    fs = 64
    length = 600  # seconds
    N = fs * length
    t = linspace(0, length, num=N, endpoint=False)

    # Generate a sinusoid at frequency f
    f = 0  # Hz
    a = cos(2 * pi * f * t) * 2 ** 15
    # a = exp(1j * 2 * pi * f * t) * 2 ** 15

    amp, freqs = spec_est(a, fs, ref=2 ** 15, plot=False)
    print(max(amp))
    assert np.abs(max(amp)) < 0.03


def test_spec_est_t4():
    fs = 64
    length = 600  # seconds
    N = fs * length
    t = linspace(0, length, num=N, endpoint=False)

    # Generate a sinusoid at frequency f
    f = 0  # Hz
    # a = cos(2 * pi * f * t) * 2 ** 15
    a = exp(1j * 2 * pi * f * t) * 2 ** 15

    amp, freqs = spec_est(a, fs, ref=2 ** 15, plot=False)
    print(max(amp))
    assert np.abs(max(amp)) < 0.03



import random
import sys
import test.iio_scanner as iio_scanner
import test.rf.spec as spec
import time

import iio

import adi
import numpy as np
import pytest

target_uri_arg = None
ignore_skip = False
dev_checked = False
found_dev = False
found_devices = {}  # type: ignore
found_uris = {}  # type: ignore
URI = "ip:analog"


def pytest_addoption(parser):
    parser.addoption(
        "--error_on_filter",
        action="store_true",
        help="When device is not found generate error not skip",
    )
    parser.addoption(
        "--uri",
        action="store",
        help="Run test on device with the given uri. IP scanning will be skipped.",
    )


class BaseTestHelpers:
    devicename = "pluto"
    skipped_tests = []  # type: ignore
    classname = "adi.ad9361"
    uri = "ip:pluto.local"

    def check_skip(self):
        # Check if calling function is in skip list
        calling_func = sys._getframe(1).f_code.co_name
        global ignore_skip
        if (calling_func in self.skipped_tests) or (not self.check_dev()):
            if not ignore_skip:
                # Will skip test if board not found or calling_func in skipped_tests
                pytest.skip("Skipping")
            else:
                # Will fail if board not found or calling_func in skipped_tests
                pytest.fail("Board not found!")

    def check_dev(self):
        # Must use globals since each test is a separate class instance
        global found_devices
        global found_uris
        global target_uri_arg
        if not isinstance(self.devicename, list):
            ds = [self.devicename]
        else:
            ds = self.devicename
        dev_checked = False
        found_dev = False
        for d in ds:
            if d in found_devices:
                found_dev = found_devices[d]
                # If device was already found before, update the board interface URI
                self.uri = found_uris[d]
                dev_checked = True
                break

        if not dev_checked:
            if target_uri_arg:
                found_dev, board = iio_scanner.find_device(
                    self.devicename, target_uri_arg
                )
            else:
                found_dev, board = iio_scanner.find_device(self.devicename)

            if found_dev:
                found_devices[board.name] = found_dev
                found_uris[board.name] = board.uri
                self.uri = board.uri
            else:
                for d in ds:
                    found_devices[d] = False
                    found_uris[d] = ""
        return found_dev

    def dev_interface(self, val, attr, tol):
        sdr = eval(self.classname + "(uri='" + self.uri + "')")
        # Check hardware
        setattr(sdr, attr, val)
        rval = float(getattr(sdr, attr))
        del sdr
        if not isinstance(val, str):
            if abs(val - rval) > tol:
                print("Failed to set: " + attr)
                print("Set: " + str(val))
                print("Got: " + str(rval))
            return abs(val - rval)
        else:
            return val == str(rval)


class BoardInterface(BaseTestHelpers):
    def __init__(self, classname, devicename):
        self.classname = classname
        self.devicename = devicename
        self.uri = ""
        self.check_skip()


def attribute_single_value(classname, devicename, attr, start, stop, step, tol):
    bi = BoardInterface(classname, devicename)
    # Pick random number in operational range
    numints = int((stop - start) / step)
    ind = random.randint(0, numints + 1)
    val = start + step * ind
    # Check hardware
    assert bi.dev_interface(val, attr, tol) <= tol


def attribute_single_value_str(classname, devicename, attr, val, tol):
    bi = BoardInterface(classname, devicename)
    # Check hardware
    assert bi.dev_interface(str(val), attr, tol) <= tol


def attribute_single_value_pow2(classname, devicename, attr, max_pow, tol):
    bi = BoardInterface(classname, devicename)
    # Pick random number in operational range
    nums = []
    for k in range(0, max_pow):
        nums.append(2 ** k)
    ind = random.randint(0, len(nums) - 1)
    val = nums[ind]
    # Check hardware
    assert bi.dev_interface(val, attr, tol) <= tol


def dma_rx(classname, devicename, channel):
    bi = BoardInterface(classname, devicename)
    sdr = eval(bi.classname + "(uri='" + bi.uri + "')")
    N = 2 ** 15
    if not isinstance(channel, list):
        sdr.rx_enabled_channels = [channel]
    else:
        sdr.rx_enabled_channels = channel
    sdr.rx_buffer_size = N * len(sdr.rx_enabled_channels)
    try:
        for _ in range(10):
            data = sdr.rx()
            if isinstance(data, list):
                for chan in data:
                    assert np.sum(np.abs(chan)) > 0
            else:
                assert np.sum(np.abs(data)) > 0
    except Exception as e:
        del sdr
        raise Exception(e)

    del sdr


def dma_tx(classname, devicename, channel):
    bi = BoardInterface(classname, devicename)
    sdr = eval(bi.classname + "(uri='" + bi.uri + "')")
    TXFS = 1000
    N = 2 ** 15
    ts = 1 / float(TXFS)
    t = np.arange(0, N * ts, ts)
    fc = 10000
    d = np.cos(2 * np.pi * t * fc) * 2 ** 15 * 0.5

    if not isinstance(channel, list):
        sdr.tx_enabled_channels = [channel]
    else:
        sdr.tx_enabled_channels = channel
        d = [d] * len(channel)
    sdr.tx_buffer_size = N * len(sdr.tx_enabled_channels)

    try:
        for _ in range(10):
            sdr.tx(d)
    except Exception as e:
        del sdr
        raise Exception(e)

    del sdr


def dma_loopback(classname, devicename, channel):
    bi = BoardInterface(classname, devicename)
    sdr = eval(bi.classname + "(uri='" + bi.uri + "')")
    if classname == "adi.FMComms5" and (channel in [2, 3]):
        sdr.loopback_chip_b = 1
    else:
        sdr.loopback = 1
    sdr.tx_cyclic_buffer = True
    # Create a ramp signal with different values for I and Q
    start = 0
    tx_data = np.array(range(start, 2 ** 11), dtype=np.int16)
    tx_data = tx_data << 4
    tx_data = tx_data + 1j * (tx_data * -1 - 1)
    sdr.tx_enabled_channels = [channel]
    sdr.tx_buffer_size = len(tx_data) * 2 * len(sdr.tx_enabled_channels)
    sdr.rx_enabled_channels = [channel]
    sdr.rx_buffer_size = len(tx_data) * 2 * len(sdr.rx_enabled_channels)
    try:
        sdr.tx(tx_data)
        # Flush buffers
        for _ in range(100):
            data = sdr.rx()
        # Turn off loopback (for other tests)
        if classname == "adi.FMComms5" and (channel in [2, 3]):
            sdr.loopback_chip_b = 0
        else:
            sdr.loopback = 0
    except Exception as e:
        del sdr
        raise Exception(e)
    del sdr
    # Check data
    offset = 0
    for i in range(len(data)):
        if np.real(data[i]) == start:
            for d in range(len(tx_data)):
                # print(str(data[i+offset])+" "+str(d+start))
                assert np.real(data[i + offset]) == (d + start)
                assert np.imag(data[i + offset]) == ((d + start) * -1 - 1)
                offset = offset + 1
            break


def freq_est(y, fs):
    N = len(y)
    T = 1.0 / fs
    yf = np.fft.fft(y)
    yf = np.fft.fftshift(yf)
    xf = np.linspace(-1.0 / (2.0 * T), 1.0 / (2.0 * T), N)
    # if self.do_plots:
    #     import matplotlib.pyplot as plt
    #
    #     fig, ax = plt.subplots()
    #     ax.plot(xf, 2.0 / N * np.abs(yf))
    #     plt.show()
    indx = np.argmax(np.abs(yf))
    return xf[indx]


def dds_loopback(classname, devicename, param_set, channel, frequency, scale, peak_min):
    bi = BoardInterface(classname, devicename)
    # See if we can tone using DMAs
    sdr = eval(bi.classname + "(uri='" + bi.uri + "')")
    # Set custom device parameters
    for p in param_set.keys():
        setattr(sdr, p, param_set[p])
    # Set common buffer settings
    sdr.tx_cyclic_buffer = True
    N = 2 ** 14
    # Create a sinewave waveform
    if hasattr(sdr, "sample_rate"):
        RXFS = int(sdr.sample_rate)
    else:
        RXFS = int(sdr.rx_sample_rate)

    sdr.rx_enabled_channels = [channel]
    sdr.rx_buffer_size = N * 2 * len(sdr.rx_enabled_channels)
    sdr.dds_single_tone(frequency, scale, channel)

    # Pass through SDR
    try:
        for _ in range(10):  # Wait
            data = sdr.rx()
    except Exception as e:
        del sdr
        raise Exception(e)
    del sdr
    tone_peaks, tone_freqs = spec.spec_est(data, fs=RXFS, ref=2 ** 15)
    indx = np.argmax(tone_peaks)
    diff = np.abs(tone_freqs[indx] - frequency)
    print("Peak: " + str(tone_peaks[indx]) + "@" + str(tone_freqs[indx]))
    assert (frequency * 0.01) > diff
    assert tone_peaks[indx] > peak_min


def cw_loopback(classname, devicename, channel, param_set):
    bi = BoardInterface(classname, devicename)
    # See if we can tone using DMAs
    sdr = eval(bi.classname + "(uri='" + bi.uri + "')")
    # Set custom device parameters
    for p in param_set.keys():
        setattr(sdr, p, param_set[p])
    # Set common buffer settings
    sdr.tx_cyclic_buffer = True
    N = 2 ** 14
    sdr.tx_enabled_channels = [channel]
    sdr.tx_buffer_size = N * 2 * len(sdr.tx_enabled_channels)
    sdr.rx_enabled_channels = [channel]
    sdr.rx_buffer_size = N * 2 * len(sdr.rx_enabled_channels)
    # Create a sinewave waveform
    if hasattr(sdr, "sample_rate"):
        RXFS = int(sdr.sample_rate)
    else:
        RXFS = int(sdr.rx_sample_rate)
    A = 2 ** 15
    fc = RXFS * 0.1
    ts = 1 / float(RXFS)
    t = np.arange(0, N * ts, ts)
    if sdr._complex_data:
        i = np.cos(2 * np.pi * t * fc) * A * 0.5
        q = np.sin(2 * np.pi * t * fc) * A * 0.5
        cw = i + 1j * q
    else:
        cw = np.cos(2 * np.pi * t * fc) * A * 1

    # Pass through SDR
    try:
        sdr.tx(cw)
        for _ in range(30):  # Wait to stabilize
            data = sdr.rx()
    except Exception as e:
        del sdr
        raise Exception(e)
    del sdr
    # tone_freq = freq_est(data, RXFS)
    # diff = np.abs(tone_freq - fc)
    # print("Peak: @"+str(tone_freq) )
    # assert (fc * 0.01) > diff

    tone_peaks, tone_freqs = spec.spec_est(data, fs=RXFS, ref=A)
    indx = np.argmax(tone_peaks)
    diff = np.abs(tone_freqs[indx] - fc)
    print("Peak: " + str(tone_peaks[indx]) + "@" + str(tone_freqs[indx]))
    assert (fc * 0.01) > diff
    # self.assertGreater(fc * 0.01, diff, "Frequency offset")


def t_sfdr(classname, devicename, channel, param_set, sfdr_min):
    bi = BoardInterface(classname, devicename)
    # See if we can tone using DMAs
    sdr = eval(bi.classname + "(uri='" + bi.uri + "')")
    # Set custom device parameters
    for p in param_set.keys():
        setattr(sdr, p, param_set[p])
    time.sleep(5)  # Wait for QEC to kick in
    # Set common buffer settings
    N = 2 ** 14
    sdr.tx_cyclic_buffer = True
    sdr.tx_enabled_channels = [channel]
    sdr.tx_buffer_size = N * 2 * len(sdr.tx_enabled_channels)
    sdr.rx_enabled_channels = [channel]
    sdr.rx_buffer_size = N * 2 * len(sdr.rx_enabled_channels)
    # Create a sinewave waveform
    if hasattr(sdr, "sample_rate"):
        RXFS = int(sdr.sample_rate)
    else:
        RXFS = int(sdr.rx_sample_rate)
    fc = RXFS * 0.1

    ts = 1 / float(RXFS)
    t = np.arange(0, N * ts, ts)
    i = np.cos(2 * np.pi * t * fc) * 2 ** 15 * 0.9
    q = np.sin(2 * np.pi * t * fc) * 2 ** 15 * 0.9
    iq = i + 1j * q
    # Pass through SDR
    try:
        sdr.tx(iq)
        time.sleep(3)
        for _ in range(10):  # Wait for IQ correction to stabilize
            data = sdr.rx()
    except Exception as e:
        del sdr
        raise Exception(e)
    del sdr
    val = spec.sfdr(data, plot=False)
    assert val > sfdr_min


def gain_check(
    classname, devicename, channel, param_set, dds_scale, min_rssi, max_rssi
):
    bi = BoardInterface(classname, devicename)
    # See if we can tone using DMAs
    sdr = eval(bi.classname + "(uri='" + bi.uri + "')")
    # Set custom device parameters
    for p in param_set.keys():
        setattr(sdr, p, param_set[p])
    # Enable DDSs
    if hasattr(sdr, "sample_rate"):
        fs = int(sdr.sample_rate)
    else:
        fs = int(sdr.rx_sample_rate)
    sdr.dds_single_tone(np.floor(fs * 0.1), dds_scale, channel)
    time.sleep(3)

    # Check RSSI
    rssi1 = sdr._get_iio_attr("voltage0", "rssi", False, sdr._ctrl)
    rssi2 = sdr._get_iio_attr("voltage1", "rssi", False, sdr._ctrl)
    rssi3 = sdr._get_iio_attr("voltage0", "rssi", False, sdr._ctrl_b)
    rssi4 = sdr._get_iio_attr("voltage1", "rssi", False, sdr._ctrl_b)

    if channel == 0:
        rssi = rssi1
    if channel == 1:
        rssi = rssi2
    if channel == 2:
        rssi = rssi3
    if channel == 3:
        rssi = rssi4
    print(rssi1, rssi2, rssi3, rssi4)
    assert rssi >= min_rssi
    assert rssi <= max_rssi


def cyclic_buffer(classname, devicename, channel, param_set):
    bi = BoardInterface(classname, devicename)
    # See if we can tone using DMAs
    sdr = eval(bi.classname + "(uri='" + bi.uri + "')")
    # Set custom device parameters
    for p in param_set.keys():
        setattr(sdr, p, param_set[p])
    fs = int(sdr.sample_rate)
    fc = -3000000
    N = 1024
    ts = 1 / float(fs)
    t = np.arange(0, N * ts, ts)
    i = np.cos(2 * np.pi * t * fc) * 2 ** 14
    q = np.sin(2 * np.pi * t * fc) * 2 ** 14
    iq = i + 1j * q
    sdr.tx_cyclic_buffer = True
    sdr.tx_enabled_channels = [channel]
    sdr.tx_buffer_size = N * 2 * len(sdr.tx_enabled_channels)
    sdr.tx(iq)
    sdr.tx_destroy_buffer()
    fail = False
    try:
        sdr.tx(iq)
    except Exception as e:
        fail = True
        msg = (
            "Pushing new data after destroying buffer should not fail. "
            "message: " + str(e)
        )

    # Cleanly end
    del sdr
    if fail:
        pytest.fail(msg)


def cyclic_buffer_exception(classname, devicename, channel, param_set):
    bi = BoardInterface(classname, devicename)
    # See if we can tone using DMAs
    sdr = eval(bi.classname + "(uri='" + bi.uri + "')")
    # Set custom device parameters
    for p in param_set.keys():
        setattr(sdr, p, param_set[p])
    fs = int(sdr.sample_rate)
    fc = -3000000
    N = 1024
    ts = 1 / float(fs)
    t = np.arange(0, N * ts, ts)
    i = np.cos(2 * np.pi * t * fc) * 2 ** 14
    q = np.sin(2 * np.pi * t * fc) * 2 ** 14
    iq = i + 1j * q
    sdr.tx_cyclic_buffer = True
    sdr.tx_enabled_channels = [channel]
    sdr.tx_buffer_size = N * 2 * len(sdr.tx_enabled_channels)
    try:
        sdr.tx(iq)
        sdr.tx(iq)
    except Exception as e:
        if (
            "TX buffer has been submitted in cyclic mode. "
            "To push more data the tx buffer must be destroyed first." not in str(e)
        ):
            fail = True
            msg = "Wrong exception raised, message was: " + str(e)
        else:
            fail = False
    else:
        fail = True
        msg = "ExpectedException not raised"
    # Cleanly end
    del sdr
    if fail:
        pytest.fail(msg)


def command_line_config(request):
    if request.config.getoption("--error_on_filter"):
        global ignore_skip
        ignore_skip = True

    global target_uri_arg
    target_uri_arg = request.config.getoption("--uri")
    if not target_uri_arg:
        target_uri_arg = None


#########################################
# Fixtures
@pytest.fixture()
def test_attribute_single_value(request):
    command_line_config(request)
    yield attribute_single_value


@pytest.fixture()
def test_attribute_single_value_str(request):
    command_line_config(request)
    yield attribute_single_value_str


@pytest.fixture()
def test_attribute_single_value_pow2(request):
    command_line_config(request)
    yield attribute_single_value_pow2


@pytest.fixture()
def test_dma_rx(request):
    command_line_config(request)
    yield dma_rx


@pytest.fixture()
def test_dma_tx(request):
    command_line_config(request)
    yield dma_tx


@pytest.fixture()
def test_cyclic_buffer(request):
    command_line_config(request)
    yield cyclic_buffer


@pytest.fixture()
def test_cyclic_buffer_exception(request):
    command_line_config(request)
    yield cyclic_buffer_exception


@pytest.fixture()
def test_dma_loopback(request):
    command_line_config(request)
    yield dma_loopback


@pytest.fixture()
def test_sfdr(request):
    command_line_config(request)
    yield t_sfdr


@pytest.fixture()
def test_dds_loopback(request):
    command_line_config(request)
    yield dds_loopback


@pytest.fixture()
def test_iq_loopback(request):
    command_line_config(request)
    yield cw_loopback


@pytest.fixture()
def test_cw_loopback(request):
    command_line_config(request)
    yield cw_loopback


@pytest.fixture()
def test_gain_check(request):
    command_line_config(request)
    yield gain_check



import random
import test.iio_scanner as iio_scanner

import iio

import numpy as np
import pytest
from adi import adis16460

URI = "ip:analog"
dev_checked = False
found_dev = False


def check_dev(name):
    global dev_checked
    global found_dev
    if not dev_checked:
        found_dev, board = iio_scanner.find_device(name)
        if found_dev:
            global URI
            URI = board.uri
        dev_checked = True
    return found_dev


@pytest.mark.skipif(not check_dev("packrf"), reason="ADIS16460 not connected")
def testADIS16460_SensorData():
    # See if we can get non-zero data from ADC
    global URI
    s = adis16460(uri=URI)
    data = s.rx()
    for k in data:
        s = np.sum(np.abs(k))
        assert s > 0


scalar_properties = [("sample_rate", 1, 4096, 0)]


@pytest.mark.skipif(not check_dev("packrf"), reason="ADIS16460 not connected")
@pytest.mark.parametrize("attr, start, stop, tol", scalar_properties)
def test_adis16460_attribute_single_value(attr, start, stop, tol):
    global URI
    sdr = adis16460(uri=URI)
    # Pick random number in operational range
    nums = []
    for k in range(0, 12):
        nums.append(2 ** k)
    ind = random.randint(0, len(nums) - 1)
    val = nums[ind]
    # Check hardware
    setattr(sdr, attr, val)
    rval = float(getattr(sdr, attr))
    del sdr
    if abs(val - rval) > tol:
        print("Failed to set: " + attr)
        print("Set: " + str(val))
        print("Got: " + str(rval))
    assert abs(val - rval) <= tol



from __future__ import division

from numpy import (
    absolute,
    argmax,
    argsort,
    cos,
    exp,
    floor,
    linspace,
    log10,
    multiply,
    pi,
)
from numpy.fft import fft, fftfreq, fftshift
from scipy import signal
from scipy.signal import find_peaks


def spec_est(x, fs, ref=2 ** 15, plot=False):

    N = len(x)

    # Apply window
    window = signal.kaiser(N, beta=38)
    # x = multiply(x, window)

    # Use FFT to get the amplitude of the spectrum
    ampl = 1 / N * absolute(fft(x))
    ampl = 20 * log10(ampl / ref + 10 ** -20)

    # FFT frequency bins
    freqs = fftfreq(N, 1 / fs)

    if plot:
        # Plot signal, showing how endpoints wrap from one chunk to the next
        import matplotlib.pyplot as plt

        plt.subplot(2, 1, 1)
        plt.plot(x, ".-")
        plt.plot(1, 1, "r.")  # first sample of next chunk
        plt.margins(0.1, 0.1)
        plt.xlabel("Time [s]")
        # Plot shifted data on a shifted axis
        plt.subplot(2, 1, 2)
        plt.plot(fftshift(freqs), fftshift(ampl))
        plt.margins(0.1, 0.1)
        plt.xlabel("Frequency [Hz]")
        plt.tight_layout()
        plt.show()

    return ampl, freqs


def find_peaks_cust(x, num_peaks=4):

    loc = argmax(x)
    print(x[loc : loc + 10])
    for p in range(1, num_peaks):
        l_loc = loc - 1
        r_loc = loc + 1
        l_peaks = []
        r_peaks = []
        # roll down left
        case = "up"
        while l_loc > 0:
            if case == "up":
                if x[l_loc] > x[l_loc - 1]:
                    # Data trending back up, time to find peak
                    case = "down"
            else:
                if x[l_loc] < x[l_loc - 1]:
                    # peak found
                    l_peaks.append(l_loc - 1)
                    case = "up"
            l_loc = l_loc - 1
        # roll down right
        case = "up"
        while r_loc < len(x):
            if case == "up":
                if x[r_loc - 1] < x[r_loc]:
                    # Data trending back up, time to find peak
                    case = "down"
            else:
                if x[r_loc - 1] > x[r_loc]:
                    # peak found
                    r_peaks.append(r_loc - 1)
                    case = "up"
            r_loc = r_loc + 1
        peaks = [loc] + l_peaks + r_peaks
        return peaks


def measure_peaks(x, num_peaks=4):
    peak_indxs = []
    peak_vals = []
    x_t = x.copy()
    m = min(x_t)
    for indx in range(num_peaks):
        loc = argmax(x_t)
        print(x_t[loc], x[loc], loc)
        peak_vals.append(x_t[loc])
        peak_indxs.append(loc)
        x_t[loc] = m
    return peak_vals, peak_indxs


def find_harmonics(x, freqs, num_harmonics=6, tolerance=0.01):
    vals, indxs = measure_peaks(x, num_harmonics)
    main = absolute(freqs[indxs[0]])
    main_loc = indxs[0]
    # print("Main",main)
    harmonics_locs = []
    harmonics_vals = []
    lx = len(x)
    dc_loc = floor(lx / 2)
    for indx in range(1, len(vals)):
        if absolute(indx - dc_loc) < (tolerance * lx):
            print("DC ignored", freqs[indxs[indx]])
            continue
        dif = absolute(freqs[indxs[indx]]) % main
        if dif < main * tolerance:
            harmonics_locs.append(indxs[indx])
            harmonics_vals.append(vals[indx])
            # print("Harmonic",freqs[indxs[indx]])
    return main, main_loc, harmonics_vals, harmonics_locs


def sfdr(x, fs=1, ref=2 ** 15, plot=False):
    amp, freqs = spec_est(x, fs=fs, ref=ref, plot=plot)
    amp = fftshift(amp)
    peak_indxs, _ = find_peaks(amp, distance=floor(len(x) * 0.1))

    # Sort peaks
    indxs = argsort(amp[peak_indxs])
    indxs = indxs[::-1]
    peak_indxs = peak_indxs[indxs]
    peak_vals = amp[peak_indxs]

    main = peak_vals[0]
    next = peak_vals[1]
    sfdr = absolute(main - next)

    if plot:
        import matplotlib.pyplot as plt

        plt.subplot(2, 1, 1)
        plt.plot(x, ".-")
        plt.plot(1, 1, "r.")  # first sample of next chunk
        plt.margins(0.1, 0.1)
        plt.xlabel("Time [s]")
        # Plot shifted data on a shifted axis
        plt.subplot(2, 1, 2)
        plt.plot(amp)
        plt.plot(peak_indxs[0:3], amp[peak_indxs[0:3]], "x")
        plt.margins(0.1, 0.1)
        plt.xlabel("Frequency [Hz]")
        plt.tight_layout()
        plt.show()

    return sfdr


def main():

    # import adi
    # sdr = adi.ad9361()
    # sdr.rx_buffer_size = 2**18
    # sdr.sample_rate = 10000000
    # sdr.dds_single_tone(1000000, 0.1)
    # sdr.tx_lo = 1000000000
    # sdr.rx_lo = 1000000000
    # sdr.gain_control = 'slow_attack'
    # sdr.tx_hardwaregain = -10
    # fs = sdr.sample_rate
    # for k in range(10):
    #     a = sdr.rx()[0]

    # Time is from 0 to 1 seconds, but leave off the endpoint, so
    # that 1.0 seconds is the first sample of the *next* chunk
    fs = 64
    length = 600  # seconds
    N = fs * length
    t = linspace(0, length, num=N, endpoint=False)

    # Generate a sinusoid at frequency f
    f = 10  # Hz
    # a = cos(2 * pi * f * t) * 2 ** 15
    a = exp(1j * 2 * pi * f * t) * 2 ** 15

    amp, freqs = spec_est(a, fs, ref=2 ** 15, plot=True)
    _, ml, vals, locs = find_harmonics(fftshift(amp), fftshift(freqs))
    freqs_s = fftshift(freqs)
    amp_s = fftshift(amp)

    tol = 40
    m = amp_s[ml]
    print("Main", m, freqs_s[ml])
    for p in range(len(locs)):
        if absolute(m - vals[p]) < tol:
            print("Harmonic", p + 1, "too large", vals[p], freqs_s[locs[p]])


if __name__ == "__main__":
    main()
